// dsiParser.js
import { DSIUnitNode } from './dsiUnitNode.js';
import { dsiPrefixesHTML, dsiUnitsHTML, dsiKeyWords } from './unitStrings.js';

class DSIParser {
  constructor() {
    // Default options
    this.createPerByDivision = true;
    this.maxDenominator = 10000;
    // Default wrappers for HTML output are not needed since we output directly HTML
  }

  // Singleton pattern
  static getInstance() {
    if (!DSIParser.instance) {
      DSIParser.instance = new DSIParser();
    }
    return DSIParser.instance;
  }

  parse(dsiString) {
    const warnings = [];

    // Check for non-DSI marker
    if (dsiString.length > 0 && dsiString[0] === '|') {
      warnings.push("Parsing a correctly marked non D-SI unit!");
      return {
        original: dsiString,
        tree: [[new DSIUnitNode('', dsiString.slice(1), '', false)]],
        warnings,
        nonDsiUnit: true
      };
    }

    // Replace double backslashes with single backslash
    while (dsiString.includes('\\\\')) {
      warnings.push(`Double backslash found in string, treating as one backslash: «${dsiString}»`);
      dsiString = dsiString.replace(/\\\\/g, '\\');
    }

    if (dsiString === "") {
      warnings.push("Given D-SI string is empty!");
      return {
        original: 'NULL',
        tree: [[new DSIUnitNode('', 'NULL', '', false)]],
        warnings,
        nonDsiUnit: true
      };
    }

    // Parse fractions: split at "\per" (literal string "\per")
    let modifiedStr = dsiString.replace('percent', 'prozent');
    let parts = modifiedStr.split('\\per').map(s => s.replace('prozent', 'percent'));

    // Check for empty fraction parts
    parts = parts.filter(part => {
      if (part.length === 0) {
        warnings.push(`The dsi string contains a \\per missing a numerator or denominator! Given string: ${dsiString}`);
        return false;
      }
      return true;
    });
    if (parts.length > 2) {
      warnings.push(`The dsi string contains more than one \\per, does not match specs! Given string: ${dsiString}`);
    }

    // For each fraction, parse the fractionless part.
    const tree = parts.map(part => {
      return this._parseFractionlessDsi(part, warnings);
    });

    return {
      original: dsiString,
      tree,
      warnings,
      nonDsiUnit: false
    };
  }

  _parseFractionlessDsi(dsiString, warnings) {
    let items = dsiString.split('\\');
    if (items[0] === '') {
      items.shift();
    } else {
      warnings.push(`String should start with a backslash, string given was «${dsiString}»`);
    }
    const nodes = [];
    while (items.length) {
      let prefix = '';
      let unit = '';
      let exponent = '';
      let valid = true;

      if (items.length && items[0] in dsiPrefixesHTML) {
        prefix = items.shift();
      }
      if (items.length) {
        unit = items.shift();
      } else {
        warnings.push(`This D-SI unit seems to be missing the base unit! \\${prefix}`);
        valid = false;
      }
      if (items.length && /^tothe\{[^{}]*\}$/.test(items[0])) {
        const expToken = items.shift();
        const match = expToken.match(/tothe\{([^{}]+)\}/);
        if (match) {
          exponent = match[1];
          if (isNaN(Number(exponent))) {
            warnings.push(`The exponent «${exponent}» is not a number!`);
            valid = false;
          }
        }
      }
      // If a unit token was provided but is not valid, warn and mark invalid.
      if (unit === '') {
        warnings.push(`This D-SI unit seems to be missing the base unit! \\${prefix}${exponent ? '\\tothe{' + exponent + '}' : ''}`);
        valid = false;
      } else if (!(unit in dsiUnitsHTML)) {
        warnings.push(`The identifier «${unit}» does not match any D-SI units!`);
        valid = false;
      }
      nodes.push(new DSIUnitNode(prefix, unit, exponent, valid));
    }
    return nodes;
  }
}

export default DSIParser;
