// src/dsiUnitInput.js
import { DSIUnit } from "./dsiUnit.js";
import { dsiPrefixesHTML, dsiUnitsHTML, dsiKeyWords } from "./unitStrings.js";

// Build a default suggestion list from prefixes, units, and keywords.
const defaultAllowedTokens = [
    ...Object.keys(dsiPrefixesHTML),
    ...Object.keys(dsiUnitsHTML),
    "per",
    "tothe"
].sort();

export class DSIUnitInput extends HTMLElement {
    constructor() {
        super();
        this.attachShadow({ mode: "open" });
        // Internal state.
        this._rawValue = "";
        this.liveUpdate = true;
        this.allowedTokens = defaultAllowedTokens;
        this.selectedSuggestionIndex = -1;
        this.lastAcceptedToken = ""; // Tracks the last accepted token

        // Create elements.
        this.input = document.createElement("input");
        this.input.type = "text";
        this.input.id = "dsiInput";

        this.suggestions = document.createElement("div");
        this.suggestions.className = "autocomplete-list";
        this.suggestions.style.display = "none";

        this.display = document.createElement("div"); // Rendered output on blur.
        this.display.id = "dsiDisplay";
        this.display.style.cursor = "text"; // Clickable for editing.

        // Container for input and suggestions.
        const container = document.createElement("div");
        container.style.position = "relative";
        container.appendChild(this.input);
        container.appendChild(this.suggestions);
        container.appendChild(this.display);

        // Append styles.
        const style = document.createElement("style");
        style.textContent = `
      #dsiInput {
        width: 400px;
        font-family: monospace;
        padding: 4px;
        box-sizing: border-box;
      }
      .autocomplete-list {
        position: absolute;
        border: 1px solid #ccc;
        background: #fff;
        max-height: 150px;
        overflow-y: auto;
        width: 400px;
        z-index: 100;
      }
      .autocomplete-item {
        padding: 4px;
        cursor: pointer;
      }
      .autocomplete-item.selected {
        background-color: #bde4ff;
      }
      #dsiDisplay {
        margin-top: 5px;
        padding: 4px;
        border: 1px solid #ccc;
        display: none;
        width: 400px;
        font-family: monospace;
      }
    `;
        this.shadowRoot.append(style, container);

        // Bind event handlers.
        this.onInput = this.onInput.bind(this);
        this.onKeyDown = this.onKeyDown.bind(this);
        this.onBlur = this.onBlur.bind(this);
        this.onFocus = this.onFocus.bind(this);
    }

    static get observedAttributes() {
        return ["suggestions-list"];
    }

    attributeChangedCallback(name, oldValue, newValue) {
        if (name === "suggestions-list") {
            if (newValue) {
                this.allowedTokens = newValue.split(",").map(s => s.trim());
            } else {
                this.allowedTokens = defaultAllowedTokens;
            }
        }
    }

    connectedCallback() {
        const suggestionsAttr = this.getAttribute("suggestions-list");
        if (suggestionsAttr) {
            this.allowedTokens = suggestionsAttr.split(",").map(s => s.trim());
        }
        this.input.addEventListener("input", this.onInput);
        this.input.addEventListener("keydown", this.onKeyDown);
        this.input.addEventListener("blur", this.onBlur);
        this.input.addEventListener("focus", this.onFocus);
        this.display.addEventListener("click", this.onFocus);
    }

    disconnectedCallback() {
        this.input.removeEventListener("input", this.onInput);
        this.input.removeEventListener("keydown", this.onKeyDown);
        this.input.removeEventListener("blur", this.onBlur);
        this.input.removeEventListener("focus", this.onFocus);
        this.display.removeEventListener("click", this.onFocus);
    }

    // Determines contextual suggestions.
    getContextualSuggestions() {
        const value = this.input.value;
        // Trim the value to remove extra whitespace.
        const tokens = value.trim().split("\\").filter(t => t !== "");
        // If no complete tokens, return all.
        if (tokens.length === 0) {
            return this.allowedTokens;
        }
        // If a last accepted token exists and is a prefix, use that.
        if (this.lastAcceptedToken && (this.lastAcceptedToken in dsiPrefixesHTML)) {
            return Object.keys(dsiUnitsHTML);
        }
        // Otherwise, use the previous token from the current input.
        const prevToken = tokens.length >= 2 ? tokens[tokens.length - 2] : "<empty>";
        if (prevToken in dsiPrefixesHTML) {
            return Object.keys(dsiUnitsHTML);
        } else if (prevToken in dsiUnitsHTML) {
            if (value.includes("\\per")) {
                return this.allowedTokens.filter(token => token !== "per");
            }
            return this.allowedTokens;
        } else if (prevToken === "per" || prevToken === "tothe") {
            return this.allowedTokens;
        } else {
            return this.allowedTokens;
        }
    }

    onInput(e) {
        this._rawValue = this.input.value;
        // Auto-insert a backslash after a complete token if space is typed.
        if (this.input.value.endsWith(" ")) {
            const trimmed = this.input.value.trimEnd();
            if (!trimmed.endsWith("\\")) {
                this.input.value = trimmed + "\\";
                this._rawValue = this.input.value;
            }
        }
        this.selectedSuggestionIndex = -1;
        this.updateSuggestions();
        if (this.liveUpdate) {
            this.renderOutput();
        }
    }

    updateSuggestions() {
        const cursorPos = this.input.selectionStart;
        const value = this.input.value;
        // Do not show suggestions if caret is inside tothe braces.
        const regexTothe = /\\tothe\{([^}]*)$/;
        const substring = value.substring(0, cursorPos);
        const matchTothe = substring.match(regexTothe);
        if (matchTothe) {
            console.log("State: inside tothe braces → no suggestions");
            this.suggestions.style.display = "none";
            return;
        }
        // Capture current token (which may be empty).
        const regex = /\\([a-zA-Z]*)$/;
        const match = substring.match(regex);
        let currentToken = "";
        if (match) {
            currentToken = match[1];
        }
        // If current token exactly matches a known prefix, auto-complete it.
        if (currentToken && (currentToken in dsiPrefixesHTML)) {
            console.log("Token complete: using accepted prefix", currentToken);
            this.lastAcceptedToken = currentToken;
            // Auto-append trailing backslash if not present.
            if (!value.endsWith("\\")) {
                this.input.value = value + "\\";
                this._rawValue = this.input.value;
            }
            currentToken = "";
        }
        const tokens = value.trim().split("\\").filter(t => t !== "");
        let prevToken = tokens.length >= 2 ? tokens[tokens.length - 2] : (this.lastAcceptedToken || "<empty>");
        const contextSuggestions = this.getContextualSuggestions();
        let filtered = currentToken === ""
            ? contextSuggestions
            : contextSuggestions.filter(token => token.startsWith(currentToken));
        if (filtered.length === 0) {
            console.log("State: prevToken =", prevToken, "currentToken =", currentToken, "→ No suggestions (filtered empty)");
            this.suggestions.style.display = "none";
            return;
        }
        this.selectedSuggestionIndex = 0;
        console.log("State: prevToken =", prevToken, "currentToken =", currentToken, "contextSuggestions =", contextSuggestions, "filtered =", filtered);
        this.showSuggestions(filtered, value.length - currentToken.length);
    }

    showSuggestions(list, tokenStartIndex) {
        this.suggestions.innerHTML = "";
        if (list.length === 0) {
            this.suggestions.style.display = "none";
            return;
        }
        list.forEach((token, index) => {
            const item = document.createElement("div");
            item.textContent = "\\" + token;
            item.className = "autocomplete-item";
            if (index === this.selectedSuggestionIndex) {
                item.classList.add("selected");
            }
            item.addEventListener("mousedown", (e) => {
                e.preventDefault();
                this.acceptSuggestion(token, tokenStartIndex);
            });
            this.suggestions.appendChild(item);
        });
        this.suggestions.style.display = "block";
    }

    onKeyDown(e) {
        if (this.suggestions.style.display !== "none") {
            const items = this.suggestions.querySelectorAll(".autocomplete-item");
            if (e.key === "ArrowDown") {
                e.preventDefault();
                if (items.length > 0) {
                    this.selectedSuggestionIndex = (this.selectedSuggestionIndex + 1) % items.length;
                    this.updateSuggestionSelection();
                }
            } else if (e.key === "ArrowUp") {
                e.preventDefault();
                if (items.length > 0) {
                    this.selectedSuggestionIndex = (this.selectedSuggestionIndex - 1 + items.length) % items.length;
                    this.updateSuggestionSelection();
                }
            } else if (["Enter", "Tab", "ArrowRight"].includes(e.key)) {
                if (this.selectedSuggestionIndex !== -1) {
                    e.preventDefault();
                    const selectedItem = items[this.selectedSuggestionIndex];
                    const token = selectedItem.textContent.slice(1); // remove leading backslash
                    const tokenStartIndex = this.input.value.lastIndexOf("\\");
                    this.acceptSuggestion(token, tokenStartIndex);
                }
            }
        }
    }

    updateSuggestionSelection() {
        const items = this.suggestions.querySelectorAll(".autocomplete-item");
        items.forEach((item, idx) => {
            if (idx === this.selectedSuggestionIndex) {
                item.classList.add("selected");
            } else {
                item.classList.remove("selected");
            }
        });
    }

    acceptSuggestion(token, tokenStartIndex) {
        const value = this.input.value;
        const before = value.substring(0, tokenStartIndex);
        const after = value.substring(this.input.selectionStart);
        if (token === "tothe") {
            this.input.value = before + "\\" + token + "{}" + after;
            this._rawValue = this.input.value;
            this.lastAcceptedToken = token;
            const newPos = before.length + token.length + 2;
            this.input.setSelectionRange(newPos, newPos);
        } else {
            this.input.value = before + "\\" + token + after;
            this._rawValue = this.input.value;
            this.lastAcceptedToken = token;
        }
        this.suggestions.style.display = "none";
        this.renderOutput();
    }

    renderOutput() {
        try {
            const unit = new DSIUnit(this.input.value);
            this.display.innerHTML = unit.toHTML({ oneLine: true });
            if (unit.warnings && unit.warnings.length > 0) {
                this.display.title = unit.warnings.join("; ");
            } else {
                this.display.removeAttribute("title");
            }
        } catch (err) {
            this.display.innerHTML = `<span style="color:red;">Error: ${err.message}</span>`;
        }
    }

    onBlur(e) {
        this.renderOutput();
        this.input.style.display = "none";
        this.display.style.display = "block";
        this.suggestions.style.display = "none";
    }

    onFocus(e) {
        this.input.style.display = "block";
        this.display.style.display = "none";
        this.input.value = this._rawValue;
        this.input.focus();
        this.input.setSelectionRange(this.input.value.length, this.input.value.length);
    }

    get value() {
        return this._rawValue;
    }

    set live(value) {
        this.liveUpdate = Boolean(value);
    }
}

customElements.define("dsi-unit-input", DSIUnitInput);
